@@ -8,7 +8,6 @@
 #
 
 from __future__ import absolute_import
-from __future__ import with_statement
 
 __all__ = ['Client', 'Listener', 'Pipe']
 
@@ -21,11 +20,13 @@ import tempfile
 import itertools
 
 from . import AuthenticationError
+from . import reduction
 from ._ext import _billiard, win32
-from .compat import get_errno
-from .util import get_temp_dir, Finalize, sub_debug, debug
+from .compat import get_errno, bytes
+from .five import monotonic
 from .forking import duplicate, close
-from .compat import bytes
+from .reduction import ForkingPickler
+from .util import get_temp_dir, Finalize, sub_debug, debug
 
 try:
     WindowsError = WindowsError  # noqa
@@ -36,6 +37,9 @@ except NameError:
 # global set later
 xmlrpclib = None
 
+Connection = getattr(_billiard, 'Connection', None)
+PipeConnection = getattr(_billiard, 'PipeConnection', None)
+
 
 #
 #
@@ -60,11 +64,11 @@ if sys.platform == 'win32':
 
 
 def _init_timeout(timeout=CONNECTION_TIMEOUT):
-    return time.time() + timeout
+    return monotonic() + timeout
 
 
 def _check_timeout(t):
-    return time.time() > t
+    return monotonic() > t
 
 #
 #
@@ -81,7 +85,7 @@ def arbitrary_address(family):
         return tempfile.mktemp(prefix='listener-', dir=get_temp_dir())
     elif family == 'AF_PIPE':
         return tempfile.mktemp(prefix=r'\\.\pipe\pyc-%d-%d-' %
-                               (os.getpid(), _mmap_counter.next()))
+                               (os.getpid(), next(_mmap_counter)))
     else:
         raise ValueError('unrecognized family')
 
@@ -189,14 +193,14 @@ if sys.platform != 'win32':
         '''
         if duplex:
             s1, s2 = socket.socketpair()
-            c1 = _billiard.Connection(os.dup(s1.fileno()))
-            c2 = _billiard.Connection(os.dup(s2.fileno()))
+            c1 = Connection(os.dup(s1.fileno()))
+            c2 = Connection(os.dup(s2.fileno()))
             s1.close()
             s2.close()
         else:
             fd1, fd2 = os.pipe()
-            c1 = _billiard.Connection(fd1, writable=False)
-            c2 = _billiard.Connection(fd2, readable=False)
+            c1 = Connection(fd1, writable=False)
+            c2 = Connection(fd2, readable=False)
 
         return c1, c2
 
@@ -231,12 +235,12 @@ else:
 
         try:
             win32.ConnectNamedPipe(h1, win32.NULL)
-        except WindowsError, e:
-            if e.args[0] != win32.ERROR_PIPE_CONNECTED:
+        except WindowsError as exc:
+            if exc.args[0] != win32.ERROR_PIPE_CONNECTED:
                 raise
 
-        c1 = _billiard.PipeConnection(h1, writable=duplex)
-        c2 = _billiard.PipeConnection(h2, readable=duplex)
+        c1 = PipeConnection(h1, writable=duplex)
+        c2 = PipeConnection(h2, readable=duplex)
 
         return c1, c2
 
@@ -275,7 +279,7 @@ class SocketListener(object):
     def accept(self):
         s, self._last_accepted = self._socket.accept()
         fd = duplicate(s.fileno())
-        conn = _billiard.Connection(fd)
+        conn = Connection(fd)
         s.close()
         return conn
 
@@ -296,7 +300,7 @@ def SocketClient(address):
     while 1:
         try:
             s.connect(address)
-        except socket.error, exc:
+        except socket.error as exc:
             if get_errno(exc) != errno.ECONNREFUSED or _check_timeout(t):
                 debug('failed to connect to address %s', address)
                 raise
@@ -307,7 +311,7 @@ def SocketClient(address):
         raise
 
     fd = duplicate(s.fileno())
-    conn = _billiard.Connection(fd)
+    conn = Connection(fd)
     s.close()
     return conn
 
@@ -352,10 +356,10 @@ if sys.platform == 'win32':
             handle = self._handle_queue.pop(0)
             try:
                 win32.ConnectNamedPipe(handle, win32.NULL)
-            except WindowsError, e:
-                if e.args[0] != win32.ERROR_PIPE_CONNECTED:
+            except WindowsError as exc:
+                if exc.args[0] != win32.ERROR_PIPE_CONNECTED:
                     raise
-            return _billiard.PipeConnection(handle)
+            return PipeConnection(handle)
 
         @staticmethod
         def _finalize_pipe_listener(queue, address):
@@ -375,8 +379,8 @@ if sys.platform == 'win32':
                     address, win32.GENERIC_READ | win32.GENERIC_WRITE,
                     0, win32.NULL, win32.OPEN_EXISTING, 0, win32.NULL,
                 )
-            except WindowsError, e:
-                if e.args[0] not in (
+            except WindowsError as exc:
+                if exc.args[0] not in (
                         win32.ERROR_SEM_TIMEOUT,
                         win32.ERROR_PIPE_BUSY) or _check_timeout(t):
                     raise
@@ -388,7 +392,7 @@ if sys.platform == 'win32':
         win32.SetNamedPipeHandleState(
             h, win32.PIPE_READMODE_MESSAGE, None, None
         )
-        return _billiard.PipeConnection(h)
+        return PipeConnection(h)
 
 #
 # Authentication stuff
@@ -471,3 +475,12 @@ def XmlClient(*args, **kwds):
     global xmlrpclib
     import xmlrpclib  # noqa
     return ConnectionWrapper(Client(*args, **kwds), _xml_dumps, _xml_loads)
+
+
+if sys.platform == 'win32':
+    ForkingPickler.register(socket.socket, reduction.reduce_socket)
+    ForkingPickler.register(Connection, reduction.reduce_connection)
+    ForkingPickler.register(PipeConnection, reduction.reduce_pipe_connection)
+else:
+    ForkingPickler.register(socket.socket, reduction.reduce_socket)
+    ForkingPickler.register(Connection, reduction.reduce_connection)
