@@ -3,13 +3,22 @@ from __future__ import absolute_import
 import errno
 import os
 import sys
-import __builtin__
+
+from .five import builtins, range
+
+try:
+    import _winapi
+except ImportError:                            # pragma: no cover
+    try:
+        from billiard import win32 as _winapi  # noqa
+    except (ImportError, AttributeError):
+        _winapi = None                         # noqa
 
 if sys.version_info[0] == 3:
     bytes = bytes
 else:
     try:
-        _bytes = __builtin__.bytes
+        _bytes = builtins.bytes
     except AttributeError:
         _bytes = str
 
@@ -25,10 +34,10 @@ try:
 except AttributeError:
 
     def closerange(fd_low, fd_high):  # noqa
-        for fd in reversed(xrange(fd_low, fd_high)):
+        for fd in reversed(range(fd_low, fd_high)):
             try:
                 os.close(fd)
-            except OSError, exc:
+            except OSError as exc:
                 if exc.errno != errno.EBADF:
                     raise
 
@@ -46,3 +55,21 @@ def get_errno(exc):
         except AttributeError:
             pass
     return 0
+
+
+if _winapi:
+
+    def setblocking(handle, blocking):
+        raise NotImplementedError('setblocking not implemented on win32')
+
+else:
+    from os import O_NONBLOCK
+    from fcntl import fcntl, F_GETFL, F_SETFL
+
+    def setblocking(handle, blocking):  # noqa
+        flags = fcntl(handle, F_GETFL, 0)
+        if flags > 0:
+            fcntl(
+                handle, F_SETFL,
+                flags & (~O_NONBLOCK) if blocking else flags | O_NONBLOCK,
+            )
