@@ -14,12 +14,12 @@ import sys
 import signal
 import warnings
 
-from ._ext import Connection, PipeConnection, win32
 from pickle import load, HIGHEST_PROTOCOL
 from billiard import util, process
+from .reduction import dump
 
 __all__ = ['Popen', 'assert_spawning', 'exit',
-           'duplicate', 'close', 'ForkingPickler']
+           'duplicate', 'close']
 
 try:
     WindowsError = WindowsError  # noqa
@@ -27,6 +27,15 @@ except NameError:
     class WindowsError(Exception):  # noqa
         pass
 
+try:
+    import _winapi as win32
+except ImportError:
+    try:
+        from _billiard import win32  # noqa
+    except ImportError:
+        if sys.platform == 'win32':
+            raise
+
 W_OLD_DJANGO_LAYOUT = """\
 Will add directory %r to path! This is necessary to accommodate \
 pre-Django 1.4 layouts using setup_environ.
@@ -53,105 +62,16 @@ def assert_spawning(self):
             ' through inheritance' % type(self).__name__
         )
 
-#
-# Try making some callable types picklable
-#
-from pickle import Pickler
-
-if sys.version_info[0] == 3:
-    from copyreg import dispatch_table
-
-    class ForkingPickler(Pickler):
-        _extra_reducers = {}
-
-        def __init__(self, *args, **kwargs):
-            Pickler.__init__(self, *args, **kwargs)
-            self.dispatch_table = dispatch_table.copy()
-            self.dispatch_table.update(self._extra_reducers)
-
-        @classmethod
-        def register(cls, type, reduce):
-            cls._extra_reducers[type] = reduce
-
-    def _reduce_method(m):
-        if m.__self__ is None:
-            return getattr, (m.__class__, m.__func__.__name__)
-        else:
-            return getattr, (m.__self__, m.__func__.__name__)
-
-    class _C:
-        def f(self):
-            pass
-    ForkingPickler.register(type(_C().f), _reduce_method)
-
-else:
-
-    class ForkingPickler(Pickler):  # noqa
-        dispatch = Pickler.dispatch.copy()
-
-        @classmethod
-        def register(cls, type, reduce):
-            def dispatcher(self, obj):
-                rv = reduce(obj)
-                self.save_reduce(obj=obj, *rv)
-            cls.dispatch[type] = dispatcher
-
-    def _reduce_method(m):  # noqa
-        if m.im_self is None:
-            return getattr, (m.im_class, m.im_func.func_name)
-        else:
-            return getattr, (m.im_self, m.im_func.func_name)
-    ForkingPickler.register(type(ForkingPickler.save), _reduce_method)
-
-
-def _reduce_method_descriptor(m):
-    return getattr, (m.__objclass__, m.__name__)
-ForkingPickler.register(type(list.append), _reduce_method_descriptor)
-ForkingPickler.register(type(int.__add__), _reduce_method_descriptor)
-
-try:
-    from functools import partial
-except ImportError:
-    pass
-else:
-
-    def _reduce_partial(p):
-        return _rebuild_partial, (p.func, p.args, p.keywords or {})
-
-    def _rebuild_partial(func, args, keywords):
-        return partial(func, *args, **keywords)
-    ForkingPickler.register(partial, _reduce_partial)
-
-
-def dump(obj, file, protocol=None):
-    ForkingPickler(file, protocol).dump(obj)
-
-#
-# Make (Pipe)Connection picklable
-#
-
-
-def reduce_connection(conn):
-    # XXX check not necessary since only registered with ForkingPickler
-    if not Popen.thread_is_spawning():
-        raise RuntimeError(
-            'By default %s objects can only be shared between processes\n'
-            'using inheritance' % type(conn).__name__
-        )
-    return type(conn), (Popen.duplicate_for_child(conn.fileno()),
-                        conn.readable, conn.writable)
-
-ForkingPickler.register(Connection, reduce_connection)
-if PipeConnection:
-    ForkingPickler.register(PipeConnection, reduce_connection)
-
 
 #
 # Unix
 #
 
 if sys.platform != 'win32':
-    import thread
+    try:
+        import thread
+    except ImportError:
+        import _thread as thread  # noqa
     import select
 
     WINEXE = False
@@ -265,9 +185,15 @@ if sys.platform != 'win32':
 #
 
 else:
-    import thread
+    try:
+        import thread
+    except ImportError:
+        import _thread as thread  # noqa
     import msvcrt
-    import _subprocess
+    try:
+        import _subprocess
+    except ImportError:
+        import _winapi as _subprocess  # noqa
 
     #
     #
@@ -287,10 +213,14 @@ else:
     def duplicate(handle, target_process=None, inheritable=False):
         if target_process is None:
             target_process = _subprocess.GetCurrentProcess()
-        return _subprocess.DuplicateHandle(
+        h = _subprocess.DuplicateHandle(
             _subprocess.GetCurrentProcess(), handle, target_process,
             0, inheritable, _subprocess.DUPLICATE_SAME_ACCESS
-        ).Detach()
+        )
+        if sys.version_info[0] < 3 or (
+                sys.version_info[0] == 3 and sys.version_info[1] < 3):
+            h.Detach()
+        return h
 
     #
     # We define a Popen class similar to the one from subprocess, but
@@ -318,8 +248,13 @@ else:
             hp, ht, pid, tid = _subprocess.CreateProcess(
                 _python_exe, cmd, None, None, 1, 0, None, None, None
             )
-            ht.Close()
-            close(rhandle)
+            if sys.version_info[0] < 3 or (
+                    sys.version_info[0] == 3 and sys.version_info[1] < 3):
+                ht.Close()
+                rhandle.Close()
+            else:
+                close(ht)
+                close(rhandle)
 
             # set attributes of self
             self.pid = pid
@@ -566,22 +501,6 @@ def get_preparation_data(name):
 
     return d
 
-    #
-    # Make (Pipe)Connection picklable
-    #
-
-    def reduce_connection(conn):
-        if not Popen.thread_is_spawning():
-            raise RuntimeError(
-                'By default %s objects can only be shared between processes\n'
-                'using inheritance' % type(conn).__name__
-            )
-        return type(conn), (Popen.duplicate_for_child(conn.fileno()),
-                            conn.readable, conn.writable)
-
-    ForkingPickler.register(Connection, reduce_connection)
-    ForkingPickler.register(PipeConnection, reduce_connection)
-
 #
 # Prepare current process
 #
@@ -659,7 +578,7 @@ def prepare(data):
             # Try to make the potentially picklable objects in
             # sys.modules['__main__'] realize they are in the main
             # module -- somewhat ugly.
-            for obj in main_module.__dict__.values():
+            for obj in list(main_module.__dict__.values()):
                 try:
                     if obj.__module__ == '__parents_main__':
                         obj.__module__ = '__main__'
