@@ -225,12 +225,13 @@ class Worker(Process):
     _job_terminated = False
 
     def __init__(self, inq, outq, synq=None, initializer=None, initargs=(),
-                 maxtasks=None, sentinel=None):
+                 maxtasks=None, sentinel=None, on_exit=None):
         assert maxtasks is None or (type(maxtasks) == int and maxtasks > 0)
         self.initializer = initializer
         self.initargs = initargs
         self.maxtasks = maxtasks
         self._shutdown = sentinel
+        self.on_exit = on_exit
         self.inq, self.outq, self.synq = inq, outq, synq
         self._make_shortcuts()
 
@@ -266,35 +267,30 @@ class Worker(Process):
             return _exit()
         sys.exit = exit
 
-        thrown = False
         pid = os.getpid()
 
         self._make_child_methods()
         self.after_fork()
-        # additional initialization before loop starts
-        self.on_loop_start(pid=pid)
+        self.on_loop_start(pid=pid)  # callback on loop start
         try:
             sys.exit(self.workloop(pid=pid))
         except Exception as exc:
-            error('Pool process error: %r', exc, exc_info=1)
-            error('outqW_fd: %r', self.outqW_fd)
-            thrown = True
-            raise
+            error('Pool process %r error: %r', self, exc, exc_info=1)
+            self._do_exit(pid, _exitcode[0], exc)
         finally:
-            # make sure finally: blocks from parent are not called.
-            if _exitcode[0] is None:
-                _exitcode[0] = EX_FAILURE if thrown else EX_OK
+            self._do_exit(pid, _exitcode[0], None)
 
-            # additional cleanup before exiting
-            self.on_loop_stop(pid=pid, exitcode=_exitcode[0])
-            os._exit(_exitcode[0])
+    def _do_exit(self, pid, exitcode, exc=None):
+        if exitcode is None:
+            exitcode = EX_FAILURE if exc else EX_OK
+
+        if self.on_exit is not None:
+            self.on_exit(pid, exitcode)
+        os._exit(exitcode)
 
     def on_loop_start(self, pid):
         pass
 
-    def on_loop_stop(self, pid=None, exitcode=None):
-        pass
-
     def terminate_controlled(self):
         self._controlled_termination = True
         self.terminate()
@@ -302,6 +298,7 @@ class Worker(Process):
     def workloop(self, debug=debug, now=monotonic, pid=None):
         pid = pid or os.getpid()
         put = self.outq.put
+        inqW_fd = self.inqW_fd
         synqW_fd = self.synqW_fd
         maxtasks = self.maxtasks
 
@@ -313,7 +310,7 @@ class Worker(Process):
             while 1:
                 if i > 60:
                     error('!!!WAIT FOR ACK TIMEOUT: job:%r fd:%r!!!',
-                          jid, self.synq._reader.fileno())
+                          jid, self.synq._reader.fileno(), exc_info=1)
                 req = _wait_for_syn()
                 if req:
                     type_, args = req
@@ -340,7 +337,7 @@ class Worker(Process):
                 except Exception:
                     result = (False, ExceptionInfo())
                 try:
-                    put((READY, (job, i, result)))
+                    put((READY, (job, i, result, inqW_fd)))
                 except Exception as exc:
                     _, _, tb = sys.exc_info()
                     try:
@@ -348,7 +345,7 @@ class Worker(Process):
                         einfo = ExceptionInfo((
                             MaybeEncodingError, wrapped, tb,
                         ))
-                        put((READY, (job, i, (False, einfo))))
+                        put((READY, (job, i, (False, einfo), inqW_fd)))
                     finally:
                         del(tb)
                 completed += 1
@@ -450,12 +447,12 @@ class PoolThread(threading.Thread):
             return self.body()
         except RestartFreqExceeded as exc:
             error("Thread %r crashed: %r", type(self).__name__, exc,
-                  exc_info=True)
+                  exc_info=1)
             _kill(os.getpid(), signal.SIGTERM)
             sys.exit()
         except Exception as exc:
             error("Thread %r crashed: %r", type(self).__name__, exc,
-                  exc_info=True)
+                  exc_info=1)
             os._exit(1)
 
     def start(self, *args, **kwargs):
@@ -547,7 +544,7 @@ class TaskHandler(PoolThread):
                     continue
                 break
             except Exception as exc:
-                print("Task Handler ERROR: %r" % (exc, ))
+                error('Task Handler ERROR: %r', exc, exc_info=1)
                 break
         else:
             debug('task handler got sentinel')
@@ -701,7 +698,8 @@ class TimeoutHandler(PoolThread):
 class ResultHandler(PoolThread):
 
     def __init__(self, outqueue, get, cache, poll,
-                 join_exited_workers, putlock, restart_state, check_timeouts):
+                 join_exited_workers, putlock, restart_state,
+                 check_timeouts, on_job_ready):
         self.outqueue = outqueue
         self.get = get
         self.cache = cache
@@ -712,6 +710,7 @@ class ResultHandler(PoolThread):
         self._it = None
         self._shutdown_complete = False
         self.check_timeouts = check_timeouts
+        self.on_job_ready = on_job_ready
         self._make_methods()
         super(ResultHandler, self).__init__()
 
@@ -723,16 +722,19 @@ class ResultHandler(PoolThread):
         cache = self.cache
         putlock = self.putlock
         restart_state = self.restart_state
+        on_job_ready = self.on_job_ready
 
         def on_ack(job, i, time_accepted, pid, synqW_fd):
+            restart_state.R = 0
             try:
                 cache[job]._ack(i, time_accepted, pid, synqW_fd)
             except (KeyError, AttributeError):
                 # Object gone or doesn't support _ack (e.g. IMAPIterator).
                 pass
 
-        def on_ready(job, i, obj):
-            restart_state.R = 0
+        def on_ready(job, i, obj, inqW_fd):
+            if on_job_ready is not None:
+                on_job_ready(job, i, obj, inqW_fd)
             try:
                 item = cache[job]
             except KeyError:
@@ -885,7 +887,9 @@ class Pool(object):
                  semaphore=None,
                  putlocks=False,
                  allow_restart=False,
-                 synack=False):
+                 synack=False,
+                 on_process_exit=None,
+                 **kwargs):
         self.synack = synack
         self._setup_queues()
         self._taskqueue = Queue()
@@ -896,6 +900,7 @@ class Pool(object):
         self._maxtasksperchild = maxtasksperchild
         self._initializer = initializer
         self._initargs = initargs
+        self._on_process_exit = on_process_exit
         self.lost_worker_timeout = lost_worker_timeout or LOST_WORKER_TIMEOUT
         self.on_process_up = on_process_up
         self.on_process_down = on_process_down
@@ -920,6 +925,10 @@ class Pool(object):
                 not isinstance(initializer, Callable):
             raise TypeError('initializer must be a callable')
 
+        if on_process_exit is not None and \
+                not isinstance(on_process_exit, Callable):
+            raise TypeError('on_process_exit must be callable')
+
         self._pool = []
         self._poolctrl = {}
         self.putlocks = putlocks
@@ -976,9 +985,12 @@ class Pool(object):
             self._outqueue, self._quick_get, self._cache,
             self._poll_result, self._join_exited_workers,
             self._putlock, self.restart_state, self.check_timeouts,
-            **extra_kwargs
+            self.on_job_ready, **extra_kwargs
         )
 
+    def on_job_ready(self, job, i, obj, inqW_fd):
+        pass
+
     def _help_stuff_finish_args(self):
         return self._inqueue, self._task_handler, self._pool
 
@@ -1008,7 +1020,7 @@ class Pool(object):
         inq, outq, synq = self.get_process_queues()
         w = self.Worker(
             inq, outq, synq, self._initializer, self._initargs,
-            self._maxtasksperchild, sentinel,
+            self._maxtasksperchild, sentinel, self._on_process_exit,
         )
         self._pool.append(w)
         self._process_register_queues(w, (inq, outq, synq))
@@ -1056,15 +1068,19 @@ class Pool(object):
                 exitcodes[worker.pid] = worker.exitcode
                 if worker.exitcode not in (EX_OK, EX_RECYCLE) and \
                         not getattr(worker, '_controlled_termination', False):
-                    error('Process %r pid:%r exited with exitcode %r' % (
-                        worker.name, worker.pid, worker.exitcode))
+                    error(
+                        'Process %r pid:%r exited with exitcode %r',
+                        worker.name, worker.pid, worker.exitcode, exc_info=1,
+                    )
                 self.process_flush_queues(worker)
                 del self._pool[i]
                 del self._poolctrl[worker.pid]
         if cleaned:
+            all_pids = [w.pid for w in self._pool]
             for job in list(self._cache.values()):
                 acked_by_gone = next(
-                    (pid for pid in job.worker_pids() if pid in cleaned),
+                    (pid for pid in job.worker_pids()
+                     if pid in cleaned or pid not in all_pids),
                     None
                 )
                 # already accepted by process
@@ -1332,7 +1348,8 @@ class Pool(object):
                     callback=None, error_callback=None, accept_callback=None,
                     timeout_callback=None, waitforslot=None,
                     soft_timeout=None, timeout=None, lost_worker_timeout=None,
-                    callbacks_propagate=()):
+                    callbacks_propagate=(),
+                    correlation_id=None):
         '''
         Asynchronous equivalent of `apply()` method.
 
@@ -1370,6 +1387,7 @@ class Pool(object):
                 on_timeout_cancel=self.on_timeout_cancel,
                 callbacks_propagate=callbacks_propagate,
                 send_ack=self.send_ack if self.synack else None,
+                correlation_id=correlation_id,
             )
             if timeout or soft_timeout:
                 # start the timeout handler thread when required.
@@ -1562,7 +1580,9 @@ class ApplyResult(object):
                  timeout_callback=None, error_callback=None, soft_timeout=None,
                  timeout=None, lost_worker_timeout=LOST_WORKER_TIMEOUT,
                  on_timeout_set=None, on_timeout_cancel=None,
-                 callbacks_propagate=(), send_ack=None):
+                 callbacks_propagate=(), send_ack=None,
+                 correlation_id=None):
+        self.correlation_id = correlation_id
         self._mutex = Lock()
         self._event = threading.Event()
         self._job = next(job_counter)
@@ -1629,8 +1649,8 @@ class ApplyResult(object):
             except self._callbacks_propagate:
                 raise
             except Exception as exc:
-                error("Pool callback raised exception: %r", exc,
-                      exc_info=True)
+                error('Pool callback raised exception: %r', exc,
+                      exc_info=1)
 
     def _set(self, i, obj):
         with self._mutex:
